/*  $Id: thread_win32.cpp,v 1.13 2004/02/28 14:24:38 sphair Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
**
*/

#include "Core/precomp.h"

#include "API/Core/System/thread.h"
#include "API/Core/System/error.h"
#include "API/Core/System/crash_reporter.h"
#include "thread_win32.h"

/////////////////////////////////////////////////////////////////////////////
// CL_Thread_Helper win32 implementation:

class CL_Thread_Helper : public CL_Runnable
{
public:
	CL_Thread_Helper(int (*func)(void*), void *value)
		: func(func), value(value)
	{
	}

	virtual void run()
	{
		func(value);
	}

private:
	int (*func)(void*);

	void *value;
};

/////////////////////////////////////////////////////////////////////////////
// CL_Thread win32 implementation:

CL_Thread::CL_Thread()
: impl(0)
{
}

CL_Thread::CL_Thread(int (*func)(void*), void *value)
: impl(new CL_Thread_Generic)
{
	impl->runnable = new CL_Thread_Helper(func, value);
	impl->delete_runnable = true;
	impl->thread_handle = NULL;
	impl->ref_count = 1;
}

CL_Thread::CL_Thread(CL_Runnable *runnable, bool delete_runnable)
: impl(new CL_Thread_Generic)
{
	impl->runnable = runnable;
	impl->delete_runnable = delete_runnable;
	impl->thread_handle = NULL;
	impl->ref_count = 1;
}

CL_Thread::CL_Thread(const CL_Thread &copy)
: impl(copy.impl)
{
	if (impl) impl->ref_count++;
}

CL_Thread::~CL_Thread()
{
	if (impl)
	{
		impl->ref_count--;
		if (impl->ref_count == 0)
		{
			if (impl->thread_handle != NULL) terminate();
			if (impl->delete_runnable) delete impl->runnable;
			delete impl;
		}
	}
}

unsigned long __stdcall CL_Thread_Generic::func_proxy(void *arg)
{
#ifndef _DEBUG
	// Create minidumps when not in debug mode.
	CL_CrashReporter crash_reporter;
#endif

	CL_Thread_Generic *self = (CL_Thread_Generic *) arg;
	self->runnable->run();

	return 0;
}

CL_Thread &CL_Thread::operator =(const CL_Thread &copy)
{
	if (impl)
	{
		impl->ref_count--;
		if (impl->ref_count == 0)
		{
			if (impl->thread_handle != NULL) terminate();
			if (impl->delete_runnable) delete impl->runnable;
			delete impl;
		}
	}

	impl = copy.impl;
	if (impl) impl->ref_count++;
	return *this;
}

void CL_Thread::start()
{
	impl->thread_handle = CreateThread(
		NULL,
		0,
		impl->func_proxy,
		impl,
		0,
		&impl->thread_id);

	if (impl->thread_handle == NULL)
		throw CL_Error("Failed to create thread");
}

void CL_Thread::terminate()
{
	TerminateThread(impl->thread_handle, 0);
	CloseHandle(impl->thread_handle);
	impl->thread_handle = NULL;
}

void CL_Thread::wait()
{
	if (impl->thread_handle == NULL) return;

	WaitForSingleObject(impl->thread_handle, INFINITE);
	CloseHandle(impl->thread_handle);
	impl->thread_handle = NULL;
}

void CL_Thread::set_priority(EThreadPriority priority)
{
	if (impl->thread_handle == NULL) return;

	int prio = THREAD_PRIORITY_NORMAL;
	switch (priority)
	{
	case cl_priority_above_normal:  prio = THREAD_PRIORITY_ABOVE_NORMAL;  break;
	case cl_priority_below_normal:  prio = THREAD_PRIORITY_BELOW_NORMAL;  break;
	case cl_priority_highest:       prio = THREAD_PRIORITY_HIGHEST;       break;
	case cl_priority_idle:          prio = THREAD_PRIORITY_IDLE;          break;
	case cl_priority_lowest:        prio = THREAD_PRIORITY_LOWEST;        break;
	case cl_priority_normal:        prio = THREAD_PRIORITY_NORMAL;        break;
	case cl_priority_time_critical: prio = THREAD_PRIORITY_TIME_CRITICAL; break;
	}

	BOOL result = SetThreadPriority(impl->thread_handle, prio);
	if (result == FALSE) throw CL_Error("Failed to set thread priority");
}

int CL_Thread::get_current_id()
{
	return GetCurrentThreadId();
}
