/*  $Id: input_device_x11keyboard.cpp,v 1.2 2004/04/18 21:17:48 sphair Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <cstdio>
#include "API/Display/input_device.h"
#include "API/Display/input_event.h"
#include "API/Display/keys.h"
#include "input_device_x11keyboard.h"
#include "../../GL/GLX/display_window_opengl.h"

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_X11Keyboard construction:

CL_InputDevice_X11Keyboard::CL_InputDevice_X11Keyboard(CL_DisplayWindow_OpenGL *owner) :
	owner(owner)
{
	type = CL_InputDevice::keyboard;

	slot_xevent = owner->sig_xevent.connect(
		this, &CL_InputDevice_X11Keyboard::on_xevent);
}

CL_InputDevice_X11Keyboard::~CL_InputDevice_X11Keyboard()
{
}

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_X11Keyboard attributes:


/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_X11Keyboard operations:

std::string CL_InputDevice_X11Keyboard::get_key_name(int virtual_key) const
{
	// Look up key name:
	char *name = XKeysymToString(virtual_key);
	if (name && name[0] != 0) return name;

	// Unknown. Return something at least :)
	char buffer[256];
	snprintf(buffer, 256, "Unknown %d", virtual_key);
	return buffer;
}

bool CL_InputDevice_X11Keyboard::get_keycode(int keysym) const
{
	// Ignore all key events when we don't have focus
	if(!owner->has_focus())
		return false;

	char keyboard_state[32];

	KeyCode code = XKeysymToKeycode(owner->get_display(), keysym);
	XQueryKeymap(owner->get_display(), keyboard_state);

	return keyboard_state[code/8] & (1 << code%8);
}

float CL_InputDevice_X11Keyboard::get_axis(int index) const
{
	return 0.0f;
}

std::string CL_InputDevice_X11Keyboard::get_name() const
{
	return "System Keyboard";
}

int CL_InputDevice_X11Keyboard::get_axis_count() const
{
	return 0;
}

int CL_InputDevice_X11Keyboard::get_button_count() const
{
	return -1;
}

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_X11Keyboard implementation:

void CL_InputDevice_X11Keyboard::on_xevent(XEvent &event)
{
	// Only handle keyboard events.
	if (event.type != KeyPress && event.type != KeyRelease) return;

	// Figure out what key it was:
	KeySym sym = XLookupKeysym(&event.xkey, 0);

	// set state of modifier keys
	bool key_state = false;
	if (event.type == KeyPress)
		key_state = true;
	
	switch(sym)
	{
		case CL_KEY_LCONTROL: owner->left_ctrl_down = key_state; break;
		case CL_KEY_LMENU: owner->left_alt_down = key_state; break;
		case CL_KEY_LSHIFT: owner->left_shift_down = key_state; break;
		case CL_KEY_RCONTROL: owner->right_ctrl_down = key_state; break;
		case CL_KEY_RMENU: owner->right_alt_down = key_state; break;
		case CL_KEY_RSHIFT: owner->right_shift_down = key_state; break;
	}
	
	// Setup event structure for ClanLib:
	CL_InputEvent key;
	key.type = CL_InputEvent::released;
	if (event.type == KeyPress) key.type = CL_InputEvent::pressed;
	key.id = sym;
	key.device = owner->keyboard;
	key.repeat_count = 0;
	key.left_alt = owner->left_alt_down;
	key.left_ctrl = owner->left_ctrl_down;
	key.left_shift = owner->left_shift_down;
	key.right_alt = owner->right_alt_down;
	key.right_ctrl = owner->right_ctrl_down;
	key.right_shift = owner->right_shift_down;
	
	// Need to somehow get proper character sequence from X11. This code
	// wont keep proper track of deadkeys:
	char buf[11];
	buf[10] = 0;
	XLookupString(&event.xkey, buf, 10, 0, 0);
	if (strlen(buf) > 0) key.str = std::string(buf, 1);

	// Emit it.
	if (key.type == CL_InputEvent::pressed)
		owner->keyboard.sig_key_down().call(key);
	else
		owner->keyboard.sig_key_up().call(key);
}
